#include "MeshLoader.h"
using namespace std;


vector<Node>& MeshLoader::Get_Nodes() {
    return nodes;
}
vector<FiniteElement>& MeshLoader::Get_FiniteElements() {
    return finiteelements;
}
vector<BoundaryFiniteElement>& MeshLoader::Get_BFEs() {
    return bfes;
}


//!!! Все контейнеры должны возвращать ID, а не сами узлы и КЭ.

vector<int> MeshLoader::Get_FE_by_nodes(int id1, int id2, int id3) {
    vector<int> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()){
        it = find_if(it, finiteelements.end(), [id1, id2, id3](const FiniteElement &x) {
            return ((find(x.nodes_id.begin(), x.nodes_id.end(), id1) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id2) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id3) != x.nodes_id.end())
            );
        });
        if (it != finiteelements.end()) {
            ans.push_back(it->fe_id);
            it++;
        }
    }
    return ans;
}
vector<int> MeshLoader::Get_FE_by_edge(int id1, int id2) {
    vector<int> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()) {
        it = find_if(it, finiteelements.end(), [id1, id2](const FiniteElement &x)  {
            return ((find(x.nodes_id.begin(), x.nodes_id.end(), id1) != x.nodes_id.end()) &&
                    (find(x.nodes_id.begin(), x.nodes_id.end(), id2) != x.nodes_id.end())
            );
        });
        if (it != finiteelements.end()) {
            ans.push_back(it->fe_id);
            it++;
        }
    }
    return ans;
}

set<int> MeshLoader::Get_nodes_by_border(int id) {
    set <int> ans;
    auto it = bfes.begin();
    while(it != bfes.end()){
        it = find_if(it, bfes.end(), [id](const BoundaryFiniteElement &x){
            return x.border_id == id;});
        for (const auto& x : it->nodes_id)
            ans.insert(x);
        it++;
    }
    return ans;
}
vector<int>MeshLoader::Get_FE_by_material(int id) {
    vector<int> ans;
    auto it = finiteelements.begin();
    while (it != finiteelements.end()) {
        it = find_if(it, finiteelements.end(), [id](const FiniteElement& x) {
            return x.mat_id == id; });
        if (it != finiteelements.end()) {
            ans.push_back(it->fe_id);
            it++;
        }
    }
    return ans;
}

vector<int> MeshLoader::Get_BFE_by_border(int id) {
    vector<int> ans;
    auto it = bfes.begin();
    while (it != bfes.end()) {
        it = find_if(it, bfes.end(), [id](const BoundaryFiniteElement &x) {
            return x.border_id == id; });
        if (it != bfes.end()) {
            ans.push_back(it->bfe_id);
            it++;
        }
    }
    return ans;
}

void MeshLoader::Insertion() {
    //!!! Нужно завести хеш-таблицу для хеширования просмотренных ребер.
    //!!! Обходя КЭ и его ребра, нужно проверять, было ли просмотрено ребро ранее.
    //!!! Если да, то в ребре можно хранить ID третьего узла и использовать его для добавления в узлы данного КЭ.
    //!!! Если нет, то создать новый узел и захешировать ребро.

    unordered_set<Edge> inserted;
    for (auto &cur: finiteelements) {
        vector vc = cur.nodes_id;
        for (int i = 0; i < dim; i++) {
            for (int j = i + 1; j < dim; j++) {
                Edge exploring = Edge(vc[i], vc[j]);
                if (inserted.contains(exploring)) {
                    cur.nodes_id.push_back(inserted.find(exploring)->baby);
                } else {
                    int id_c = nodes.size() + 1;
                    exploring.baby = id_c;
                    cur.nodes_id.push_back(id_c);
                    Node tmp{};
                    tmp.id = id_c;
                    tmp.x = (nodes[vc[i]].x + nodes[vc[j]].x) / 2;
                    tmp.y = (nodes[vc[i]].y + nodes[vc[j]].y) / 2;
                    tmp.z = (nodes[vc[i]].z + nodes[vc[j]].z) / 2;
                    tmp.is_summit = false;
                    nodes.push_back(move(tmp));
                    inserted.insert(exploring);
                }
            }
        }
    }
    for (auto &cur: bfes) {
        vector vc = cur.nodes_id;
        for (int i = 0; i < dim; i++) {
            for (int j = i + 1; j < dim; j++) {
                Edge exploring = Edge(vc[i], vc[j]);
                if (inserted.contains(exploring)) {
                    vc.push_back(inserted.find(exploring)->baby);
                } else {
                    int id_c = nodes.size();
                    exploring.baby = id_c;
                    vc.push_back(id_c);
                    Node tmp{};
                    tmp.id = id_c;
                    tmp.x = (nodes[vc[i]].x + nodes[vc[j]].x) / 2;
                    tmp.y = (nodes[vc[i]].y + nodes[vc[j]].y) / 2;
                    tmp.z = (nodes[vc[i]].z + nodes[vc[j]].z) / 2;
                    tmp.is_summit = false;
                    nodes.push_back(move(tmp));
                    inserted.insert(exploring);
                }
            }
        }
    }
    dim = dim * 2 - 1;
}

vector <set <int>> MeshLoader::Get_neighbors(){
    vector<set<int>> ans(nodes.size() + 1);
    for (const auto &cur : finiteelements)
        for (const auto &id: cur.nodes_id)
            for (const auto &nbr: cur.nodes_id)
                if (id != nbr)
                    ans[id].insert(nbr);
    return ans;
}

void MeshLoader::printNode(const Node &node) {
    std::cout << node;
}

void MeshLoader::printFE(const FiniteElement &fe) {
    std::cout << fe;
}

void MeshLoader::printBFE(const BoundaryFiniteElement &bfe) {
    std::cout << bfe;
}

set<int> MeshLoader::getBoundaryNodesId() {
    set<int> res;
    for (const auto &cur: bfes) {
        for_each(cur.nodes_id.begin(), cur.nodes_id.end(), [&res](int id) {
            res.insert(id);
        });
    }
    return res;
}


